"""
Operational semantics for the Why language

Provides evaluation functions for the three syntactic classes,
based on a environment mapping variables to integer values.
These functions return None when an error occur somewhere (Big step semantics)
"""

from syntax import *

def eval_aexp(env: dict[str,int], exp: ArithExpr) -> int | None:
    """evaluates an arithmetic expression"""
    match exp:
        case AECst(value):
            return value
        case AEVar(var):
            if var in env:
                return env[var]
            else:
                return None
        case AEUop(uop,expr):
            if uop == Uop.OPP:
                e = eval_aexp(env, expr)
                if e is None:
                    return None
                else:
                    return -eval_aexp(env, expr)
            else:
                return None
        case AEBop(bop,left_expr,right_expr):
            l = eval_aexp(env, left_expr)
            r = eval_aexp(env, right_expr)
            if l is None or r is None:
                return None
            match bop:
                case Bop.ADD: return l + r
                case Bop.MUL: return l * r
                case Bop.DIV: return l // r if r != 0 else None
        case _: assert False

def eval_bexp(env: dict[str,int], exp:BoolExpr) -> bool | None:
    """evaluates a boolean expression"""
    match exp:
        case BEPlain(aexpr):
            e = eval_aexp(env, aexpr)
            if e is None:
                return None
            else:
                return True if e != 0 else False
        case BEEq(left_expr,right_expr):
            l = eval_aexp(env, left_expr)
            r = eval_aexp(env, right_expr)
            if l is None or r is None:
                return None
            return l == r
        case BELeq(left_expr,right_expr):
            l = eval_aexp(env, left_expr)
            r = eval_aexp(env, right_expr)
            if l is None or r is None:
                return None
            return l <= r
        case BENeg(bexpr):
            b = eval_bexp(env, bexpr)
            if b is None:
                return None
            else:
                return not b
        case _: assert False

def eval_instr(env: dict[str,int], instr: Instr) -> dict[str,int] | None:
    """evaluates an instruction"""
    match instr:
        case ISkip():
            return env.copy()
        case ISet(var,expr):
            val = eval_aexp(env, expr)
            nvEnv = env.copy()
            nvEnv[var] = val
            return nvEnv
        case ISeq(first,second):
            nvEnv1 = eval_instr(env, first)
            return eval_instr(nvEnv1, second)
        case ICond(cond,true_branch,false_branch):
            b = eval_bexp(env, cond)
            if b:
                return eval_instr(env, true_branch)
            else:
                return eval_instr(env, false_branch)
        case ILoop(cond,body):
            b = eval_bexp(env, cond)
            if b:
                nvEnv = eval_instr(env, body)
                return eval_instr(nvEnv, ILoop(cond,body))
            else:
                return env.copy()
        case _: assert False
