# pyright: strict
"""
implements a sign analysis
"""
from enum import Enum, auto
from dataclasses import dataclass
from cfg import Cfg
from iteration import Transfer, fixpoint_iteration
from syntax import *

class sign(Enum):
    SNEG = auto()
    SPOS = auto()
    Z = auto()
    NEG = auto()
    POS = auto()
    NZ = auto()
    INT = auto()

class Top:
    def __str__(self): return "TOP"

@dataclass
class BTop:
    """class used for the evaluation of boolean expressions

    BTop(False) indicates that we don't know if the result is True or False, but
    that the evaluation cannot lead to an error
    BTop(True) indicates that we neither know the result nor whether an error can occur
    """
    has_error: bool
    def __str__(self):
        if self.has_error: return "TOP (maybe error)"
        else: return "TOP (no error)"

type abstract_env = dict[str, sign]
"""mapping from variables to abstract values.

As with concrete environment, a variable not in the dictionary will
lead to an error if we try to obtain its value
"""
type state = abstract_env | None
"""abstract state is either an abstract env or bottom
"""

def sign_leq(v1: sign, v2:sign) -> bool:
    """order relation on the sign lattice"""
    if v1 == v2: return True
    match v1,v2:
        case _, sign.INT: return True
        case sign.INT, _: return False
        case sign.SNEG,(sign.NEG | sign.NZ): return True
        case sign.SPOS,(sign.POS | sign.NZ): return True
        case sign.Z,(sign.POS | sign.NEG): return True
        case sign.NEG,sign.NZ: return True
        case sign.POS,sign.NZ: return True
        case _: return False

def sign_join(v1: sign, v2:sign) -> sign:
    """computes least upper bound"""
    match v1,v2:
        case sign.INT, _: return sign.INT
        case _, sign.INT: return sign.INT
        case sign.NZ, (sign.NZ | sign.SPOS | sign.SNEG): return sign.NZ
        case sign.NZ,_: return sign.INT
        case sign.NEG, (sign.NEG | sign.SNEG | sign.Z): return sign.NEG
        case sign.NEG, _: return sign.INT
        case sign.POS, (sign.POS | sign.SPOS | sign.Z): return sign.POS
        case sign.POS,_: return sign.INT
        case sign.SNEG, sign.SNEG: return sign.SNEG
        case sign.SNEG, (sign.Z | sign.NEG): return sign.NEG
        case sign.SNEG, (sign.SPOS | sign.NZ): return sign.NZ
        case sign.SNEG, _: return sign.INT
        case sign.SPOS, sign.SPOS: return sign.SPOS
        case sign.SPOS, (sign.POS | sign.Z): return sign.POS
        case sign.SPOS, (sign.SNEG | sign.NZ): return sign.NZ
        case sign.SPOS, _: return sign.INT
        case sign.Z, sign.SNEG: return sign.NEG
        case sign.Z, sign.SPOS: return sign.POS
        case sign.Z, sign.NZ: return sign.INT
        case sign.Z, _: return v2

def sign_meet(s1: sign, s2:sign) -> sign | None:
    """computes greatest lower bound. Can be None (i.e. bottom)"""
    match s1,s2:
        case sign.INT, _: return s2
        case _, sign.INT: return s1
        case sign.SNEG, (sign.NEG | sign.NZ | sign.SNEG): return s1
        case (sign.NEG | sign.NZ | sign.SNEG), sign.SNEG: return s2
        case sign.SNEG, _: return None
        case _, sign.SNEG: return None
        case sign.SPOS, (sign.POS | sign.NZ | sign.SPOS): return s1
        case (sign.POS | sign.NZ | sign.SPOS), sign.SPOS: return s2
        case sign.SPOS, _: return None
        case _, sign.SPOS: return None
        case sign.Z, (sign.NEG | sign.POS): return s1
        case (sign.NEG | sign.POS), sign.Z: return s2
        case sign.Z, sign.Z: return s1
        case sign.Z, _: return None
        case _, sign.Z: return None
        case sign.NEG, sign.NEG: return s1
        case sign.POS, sign.POS: return s1
        case sign.NEG, sign.POS: return sign.Z
        case sign.POS, sign.NEG: return sign.Z
        case sign.NZ, sign.NEG: return sign.SNEG
        case sign.NEG, sign.NZ: return sign.SNEG
        case sign.NZ, sign.POS: return sign.SPOS
        case sign.POS, sign.NZ: return sign.SPOS
        case sign.NZ, sign.NZ: return sign.NZ

def sign_opp(s:sign) -> sign:
    """unary minus"""
    match s:
        case sign.SNEG: return sign.SPOS
        case sign.SPOS: return sign.SNEG
        case sign.POS: return sign.NEG
        case sign.NEG: return sign.POS
        case _: return s

def sign_add(s1:sign,s2:sign) -> sign:
    """addition"""
    match (s1,s2):
        case sign.INT,_: return sign.INT
        case _,sign.INT: return sign.INT
        case sign.Z,_: return s2
        case _,sign.Z: return s1
        case sign.SNEG, (sign.SNEG | sign.NEG): return sign.SNEG
        case (sign.SNEG | sign.NEG), sign.SNEG: return sign.SNEG
        case sign.NEG, sign.NEG: return sign.NEG
        case sign.SPOS, (sign.SPOS | sign.POS): return sign.SPOS
        case (sign.SPOS | sign.POS), sign.SPOS: return sign.SPOS
        case sign.POS, sign.POS: return sign.POS
        case _: return sign.INT

def sign_mul(s1: sign, s2: sign) -> sign:
    """multiplication"""
    match s1,s2:
        case sign.INT,_: return sign.INT
        case _,sign.INT: return sign.INT
        case sign.Z,_: return sign.Z
        case _,sign.Z: return sign.Z
        case sign.SPOS, _: return s2
        case _, sign.SPOS: return s1
        case sign.SNEG, _: return sign_opp(s2)
        case _,sign.SNEG: return sign_opp(s1)
        case sign.NZ, _: return sign.INT
        case _, sign.NZ: return sign.INT
        case sign.NEG, sign.NEG: return sign.POS
        case sign.NEG, sign.POS: return sign.NEG
        case sign.POS, sign.NEG: return sign.NEG
        case sign.POS, sign.POS: return sign.POS

def sign_div(s1: sign, s2: sign) -> sign | Top | None:
    """division: can lead to errors even if operands are plain sign"""
    match s1, s2:
        case _,sign.Z: return None
        case _,sign.INT: return Top()
        case _,sign.POS: return Top()
        case _,sign.NEG: return Top()
        case sign.Z,_: return sign.Z
        case sign.INT,_: return sign.INT
        case sign.NZ,_: return sign.INT
        case _,sign.NZ: return sign.INT
        case (sign.POS | sign.SPOS), sign.SPOS: return sign.POS
        case (sign.NEG | sign.SNEG), sign.SPOS: return sign.NEG
        case (sign.POS | sign.SPOS), sign.SNEG: return sign.NEG
        case (sign.NEG | sign.SNEG), sign.SNEG: return sign.POS

def eval_aexp(env: abstract_env, e: ArithExpr) -> sign | Top | None:
    """evaluate an arithmetic expression in an abstract environment
    returns None in case of error
    """
    match e:
        case AECst(value):
            if value < 0:
                return sign.NEG
            elif value > 0:
                return sign.POS
            elif value == 0:
                return sign.Z
            else:
                return None
        case AEVar(var):
            if var in env:
                return env[var]
            else :
                return None
        case AEUop(_,expr):
            s = eval_aexp(env, expr)
            if s is None:
                return None
            elif isinstance(s, Top):
                return Top()
            else:
                return sign_opp(s)
        case AEBop(bop,left_expr,right_expr):
            l = eval_aexp(env, left_expr)
            r = eval_aexp(env, right_expr)
            if l is None or r is None:
                return None
            if isinstance(l, Top) or isinstance(r, Top):
                return Top()
            match bop:
                case Bop.ADD: return sign_add(l, r)
                case Bop.MUL: return sign_mul(l, r)
                case Bop.DIV: return sign_div(l, r)
        case _: pass

def eval_bexp(env: abstract_env, e: BoolExpr) -> bool | BTop | None:
    """abstract evaluation of a boolean expression"""
    match e:
        case BEPlain(aexpr):
            res = eval_aexp(env, aexpr)
            if res is None: 
                return None
            if isinstance(res, Top) :
                return BTop(True)
            if res == sign.Z:
                return False
            if res == sign.NZ or res == sign.SNEG or res == sign.SPOS : 
                return True
            return BTop(False)
        case BEEq(left_expr,right_expr):
            l = eval_aexp(env, left_expr)
            r = eval_aexp(env, right_expr)
            if l is None or r is None:
                return None
            if isinstance(l, Top) or isinstance(r, Top) :
                return BTop(True)
            match l,r:
                case sign.SNEG, (sign.Z | sign.POS | sign.SPOS) : return False
                case (sign.Z | sign.POS | sign.SPOS), sign.SNEG : return False
                case sign.NEG, sign.SPOS : return False
                case sign.SPOS, sign.NEG : return False
                case sign.Z, sign.Z : return True
                case sign.Z, (sign.SNEG | sign.SPOS | sign.NZ) : return False
                case (sign.SNEG | sign.SPOS | sign.NZ), sign.Z : return False
                case _ : return BTop(False)
        case BELeq(left_expr,right_expr):
            l = eval_aexp(env, left_expr)
            r = eval_aexp(env, right_expr)
            if l is None or r is None:
                return None
            if isinstance(l, Top) or isinstance(r, Top) :
                return BTop(True)
            match l,r:
                case sign.SNEG, (sign.Z | sign.POS | sign.SPOS) : return True
                case sign.NEG, sign.SPOS : return True
                case sign.Z, sign.SNEG : return False
                case sign.Z, sign.SPOS : return True
                case sign.POS, sign.SNEG : return False
                case sign.SPOS, (sign.SNEG | sign.NEG | sign.Z) : return False
                case _ : return BTop(False)
        case BENeg(expr):
            b = eval_bexp(env, expr)
            if b is None:
                return None
            if isinstance(b, BTop):
                return b
            return not b
        case _: pass

def reduce_eq_sign(s: state, x: str, expr: ArithExpr) -> state:
    """reduce the value associated to x in s under the hypothesis that x == expr"""
    if s is None: return None
    v = eval_aexp(s,expr)
    if v is None: return None
    if isinstance(v,Top): return s
    res = sign_meet(s[x],v)
    if res is None: return None
    return s | { x: res }

def reduce_neq_sign(s: state, x: str, expr: ArithExpr) -> state:
    """reduce the value associated to x in s under the hypothesis that x != expr"""
    if s is None: return None
    v = eval_aexp(s,expr)
    if v is None: return None
    if isinstance(v,Top): return s
    match s[x],v:
        case sign.POS, sign.Z: return s | { x: sign.SPOS }
        case sign.NEG, sign.Z: return s | { x: sign.SNEG }
        case sign.INT, sign.Z: return s | { x: sign.NZ }
        case _: return s

def reduce_upper_bound(s: state, x: str, expr: ArithExpr) -> state:
    """reduce the value associated to x in s under the hypothesis that x <= expr"""
    if s is None: return None
    v = eval_aexp(s,expr)
    if v is None: return None
    if isinstance(v,Top): return s
    upper = sign.INT
    match v:
        case sign.NEG | sign.SNEG:
            upper = v
        case sign.Z:
            upper = sign.NEG
        case _: pass
    res = sign_meet(s[x],upper)
    if res is None: return None
    return s | { x: res }

def reduce_strict_upper_bound(s: state, x: str, expr: ArithExpr) -> state:
    """reduce the value associated to x in s under the hypothesis that x < expr"""
    if s is None: return None
    v = eval_aexp(s,expr)
    if v is None: return None
    if isinstance(v,Top): return s
    match v:
        case sign.NEG | sign.SNEG | sign.Z:
            upper = sign.SNEG
        case _:
            upper = sign.INT
    res = sign_meet(s[x],upper)
    if res is None: return None
    return s | { x: res }

def reduce_lower_bound(s: state, x: str, expr: ArithExpr) -> state:
    """reduce the value associated to x in s under the hypothesis that x >= expr"""
    if s is None: return None
    v = eval_aexp(s,expr)
    if v is None: return None
    if isinstance(v,Top): return s
    lower = sign.INT
    match v:
        case sign.POS | sign.SPOS:
            lower = v
        case sign.Z:
            lower = sign.POS
        case _: pass
    res = sign_meet(s[x],lower)
    if res is None: return None
    return s | { x: res }

def reduce_strict_lower_bound(s: state, x: str, expr: ArithExpr) -> state:
    """reduce the value associated to x in s under the hypothesis that x > expr"""
    if s is None: return s
    v = eval_aexp(s,expr)
    if v is None: return None
    if isinstance(v,Top): return s
    match v:
        case sign.POS | sign.SPOS | sign.Z:
            lower = sign.SPOS
        case _:
            lower = sign.INT
    res = sign_meet(s[x],lower)
    if res is None: return None
    return s | { x: res }

def reduce_state(s: state,c: BoolExpr) -> state:
    if s is None: return s
    match c:
        case BEPlain(aexpr):
            if isinstance(aexpr, AEVar):
                return reduce_state(s, BENeg(BEEq(aexpr, AECst(0))))
            else: return s
        case BEEq(left_expr,right_expr):
            if isinstance(left_expr, AEVar) and isinstance(right_expr, AEVar):
                return reduce_eq_sign(reduce_eq_sign(s, left_expr.var, right_expr), right_expr.var, left_expr)
            if isinstance(left_expr, AEVar):
                return reduce_eq_sign(s, left_expr.var, right_expr)
            if isinstance(right_expr, AEVar):
                return reduce_eq_sign(s, right_expr.var, left_expr)
            return s
        case BELeq(left_expr,right_expr):
            if isinstance(left_expr, AEVar) and isinstance(right_expr, AEVar):
                return reduce_lower_bound(reduce_upper_bound(s, left_expr.var, right_expr), right_expr.var, left_expr)
            if isinstance(left_expr, AEVar):
                return reduce_upper_bound(s, left_expr.var, right_expr)
            if isinstance(right_expr, AEVar):
                return reduce_lower_bound(s, right_expr.var, left_expr)
            return s
        case BENeg(expr):
            match expr:
                case BEPlain(aexpr): 
                    if isinstance(aexpr, AEVar):
                        return reduce_eq_sign(s, aexpr.var, AECst(0))
                    else: return s
                case BEEq(left_expr,right_expr):
                    if isinstance(left_expr, AEVar) and isinstance(right_expr, AEVar):
                        return reduce_neq_sign(reduce_neq_sign(s, left_expr.var, right_expr), right_expr.var, left_expr)
                    if isinstance(left_expr, AEVar):
                        return reduce_neq_sign(s, left_expr.var, right_expr)
                    if isinstance(right_expr, AEVar):
                        return reduce_neq_sign(s, right_expr.var, left_expr)
                    return s
                case BELeq(left_expr,right_expr):
                    if isinstance(left_expr, AEVar) and isinstance(right_expr, AEVar):
                        return reduce_strict_upper_bound(reduce_strict_lower_bound(s, left_expr.var, right_expr), right_expr.var, left_expr)
                    if isinstance(left_expr, AEVar):
                        return reduce_strict_lower_bound(s, left_expr.var, right_expr)
                    if isinstance(right_expr, AEVar):
                        return reduce_strict_upper_bound(s, right_expr.var, left_expr)
                    return s
                case BENeg(e):
                    return reduce_state(s, e)
                case _: return s
        case _: return s

class Sign_interp(Transfer[state]):
    variables: frozenset[str]
    def __init__(self,instr: Instr):
        self.variables = variables_of_instr(instr)
    def bottom(self) -> state:
        return None
    def init_state(self) -> state:
        return dict.fromkeys(self.variables, sign.INT)
    def join(self,s1:state,s2:state) -> state:
        if s1 is None: return s2
        if s2 is None: return s1
        res: abstract_env = {}
        for var in self.variables:
            v1 = s1[var]
            v2 = s2[var]
            res[var] = sign_join(v1,v2)
        return res

    def included(self,s1: state,s2: state) -> bool:
        if s1 is None: return True
        if s2 is None: return False
        for var in self.variables:
            if not sign_leq(s1[var], s2[var]): return False
        return True

    def tr_skip(self,s: state) -> state:
        return s

    def tr_set(self,s: state,v: str,e: ArithExpr) -> state:
        if s is None:
            return None
        return reduce_state(s, BEEq(AEVar(v), e))
    
    def tr_test(self,s: state,c: BoolExpr) -> state:
        if s is None:
            return None
        return reduce_state(s, c)

    def tr_err(self,s: state,e: Expr) -> state:
        if s is None: return s
        if isinstance(e,ArithExpr):
            return reduce_state(s, BEPlain(e))
        if isinstance(e,BoolExpr):
            return reduce_state(s, e)

def analyze(i: Instr) -> None:
    cfg = Cfg(i)
    res = fixpoint_iteration(Sign_interp(i), cfg)
    for node in cfg.g.nodes:
        print(f"State at {node}:")
        s = res[node]
        if s is not None:
            for (v, s) in s.items():
                print(f"  {v}: {s}")
